/**
 * Borrow from https://github.com/excalidraw/excalidraw/blob/master/packages/excalidraw/change.ts#L399
 */
var __rest = (this && this.__rest) || function (s, e) {
    var t = {};
    for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p) && e.indexOf(p) < 0)
        t[p] = s[p];
    if (s != null && typeof Object.getOwnPropertySymbols === "function")
        for (var i = 0, p = Object.getOwnPropertySymbols(s); i < p.length; i++) {
            if (e.indexOf(p[i]) < 0 && Object.prototype.propertyIsEnumerable.call(s, p[i]))
                t[p[i]] = s[p[i]];
        }
    return t;
};
import { randomInteger, Name, Visibility, FillSolid, Stroke, Text, Rect, isGradient, FillGradient, Circle, Ellipse, } from '@infinite-canvas-tutorial/ecs';
import { isNil } from '@antv/util';
import { Delta } from './Delta';
import { newElementWith } from './Snapshot';
import { createOrSetComponent, getUpdatedTimestamp, removeComponent, } from '../utils';
export class ElementsChange {
    static empty() {
        return ElementsChange.create(new Map(), new Map(), new Map(), undefined);
    }
    static stripIrrelevantProps(partial) {
        const { id, updated, version, versionNonce } = partial, strippedPartial = __rest(partial, ["id", "updated", "version", "versionNonce"]);
        return strippedPartial;
    }
    /**
     * Calculates the `Delta`s between the previous and next set of elements.
     *
     * @param prevElements - Map representing the previous state of elements.
     * @param nextElements - Map representing the next state of elements.
     *
     * @returns `ElementsChange` instance representing the `Delta` changes between the two sets of elements.
     */
    static calculate(prevElements, nextElements, api) {
        if (prevElements === nextElements) {
            return ElementsChange.empty();
        }
        const added = new Map();
        const removed = new Map();
        const updated = new Map();
        // this might be needed only in same edge cases, like during collab, when `isDeleted` elements get removed or when we (un)intentionally remove the elements
        for (const prevElement of prevElements.values()) {
            const nextElement = nextElements.get(prevElement.id);
            if (!nextElement) {
                const deleted = Object.assign(Object.assign({}, prevElement), { isDeleted: false });
                const inserted = { isDeleted: true };
                const delta = Delta.create(deleted, inserted, ElementsChange.stripIrrelevantProps);
                removed.set(prevElement.id, delta);
            }
        }
        for (const nextElement of nextElements.values()) {
            const prevElement = prevElements.get(nextElement.id);
            if (!prevElement) {
                const deleted = { isDeleted: true };
                const inserted = Object.assign(Object.assign({}, nextElement), { isDeleted: false });
                const delta = Delta.create(deleted, inserted, ElementsChange.stripIrrelevantProps);
                added.set(nextElement.id, delta);
                continue;
            }
            if (prevElement.versionNonce !== nextElement.versionNonce) {
                const delta = Delta.calculate(prevElement, nextElement, ElementsChange.stripIrrelevantProps);
                if (
                // making sure we don't get here some non-boolean values (i.e. undefined, null, etc.)
                typeof prevElement.isDeleted === 'boolean' &&
                    typeof nextElement.isDeleted === 'boolean' &&
                    prevElement.isDeleted !== nextElement.isDeleted) {
                    // notice that other props could have been updated as well
                    if (prevElement.isDeleted && !nextElement.isDeleted) {
                        added.set(nextElement.id, delta);
                    }
                    else {
                        removed.set(nextElement.id, delta);
                    }
                    continue;
                }
                // making sure there are at least some changes
                if (!Delta.isEmpty(delta)) {
                    updated.set(nextElement.id, delta);
                }
            }
        }
        return ElementsChange.create(added, removed, updated, api);
    }
    /**
     * Check for visible changes regardless of whether they were removed, added or updated.
     */
    static checkForVisibleDifference(element, partial) {
        if (element.isDeleted && partial.isDeleted !== false) {
            // when it's deleted and partial is not false, it cannot end up with a visible change
            return false;
        }
        if (element.isDeleted && partial.isDeleted === false) {
            // when we add an element, it results in a visible change
            return true;
        }
        if (element.isDeleted === false && partial.isDeleted) {
            // when we remove an element, it results in a visible change
            return true;
        }
        // check for any difference on a visible element
        return Delta.isRightDifferent(element, partial);
    }
    static applyDelta(element, delta, flags = {
        // by default we don't care about about the flags
        containsVisibleDifference: true,
        containsZindexDifference: true,
    }) {
        const directlyApplicablePartial = __rest(delta.inserted, []);
        // if (
        //   delta.deleted.boundElements?.length ||
        //   delta.inserted.boundElements?.length
        // ) {
        //   const mergedBoundElements = Delta.mergeArrays(
        //     element.boundElements,
        //     delta.inserted.boundElements,
        //     delta.deleted.boundElements,
        //     (x) => x.id,
        //   );
        //   Object.assign(directlyApplicablePartial, {
        //     boundElements: mergedBoundElements,
        //   });
        // }
        // if (isImageElement(element)) {
        //   const _delta = delta as Delta<ElementPartial<ExcalidrawImageElement>>;
        //   // we want to override `crop` only if modified so that we don't reset
        //   // when undoing/redoing unrelated change
        //   if (_delta.deleted.crop || _delta.inserted.crop) {
        //     Object.assign(directlyApplicablePartial, {
        //       // apply change verbatim
        //       crop: _delta.inserted.crop ?? null,
        //     });
        //   }
        // }
        if (!flags.containsVisibleDifference) {
            // strip away fractional as even if it would be different, it doesn't have to result in visible change
            const { index } = directlyApplicablePartial, rest = __rest(directlyApplicablePartial, ["index"]);
            const containsVisibleDifference = ElementsChange.checkForVisibleDifference(element, rest);
            flags.containsVisibleDifference = containsVisibleDifference;
        }
        if (!flags.containsZindexDifference) {
            flags.containsZindexDifference =
                delta.deleted.index !== delta.inserted.index;
        }
        return newElementWith(element, directlyApplicablePartial);
    }
    static create(added, removed, updated, api, options = { shouldRedistribute: false }) {
        let change;
        if (options.shouldRedistribute) {
            const nextAdded = new Map();
            const nextRemoved = new Map();
            const nextUpdated = new Map();
            const deltas = [...added, ...removed, ...updated];
            for (const [id, delta] of deltas) {
                if (this.satisfiesAddition(delta)) {
                    nextAdded.set(id, delta);
                }
                else if (this.satisfiesRemoval(delta)) {
                    nextRemoved.set(id, delta);
                }
                else {
                    nextUpdated.set(id, delta);
                }
            }
            change = new ElementsChange(nextAdded, nextRemoved, nextUpdated, api);
        }
        else {
            change = new ElementsChange(added, removed, updated, api);
        }
        return change;
    }
    constructor(added, removed, updated, api) {
        this.added = added;
        this.removed = removed;
        this.updated = updated;
        this.api = api;
    }
    inverse() {
        const inverseInternal = (deltas) => {
            const inversedDeltas = new Map();
            for (const [id, delta] of deltas.entries()) {
                inversedDeltas.set(id, Delta.create(delta.inserted, delta.deleted));
            }
            return inversedDeltas;
        };
        const added = inverseInternal(this.added);
        const removed = inverseInternal(this.removed);
        const updated = inverseInternal(this.updated);
        // notice we inverse removed with added not to break the invariants
        return ElementsChange.create(removed, added, updated, this.api);
    }
    /**
     * Update delta/s based on the existing elements.
     *
     * @param elements current elements
     * @param modifierOptions defines which of the delta (`deleted` or `inserted`) will be updated
     * @returns new instance with modified delta/s
     */
    applyLatestChanges(elements) {
        const modifier = (element) => (partial) => {
            // (element: OrderedSerializedNode) => (partial: ElementPartial) => {
            const latestPartial = {};
            for (const key of Object.keys(partial)) {
                // do not update following props:
                // - `boundElements`, as it is a reference value which is postprocessed to contain only deleted/inserted keys
                switch (key) {
                    // case 'boundElements':
                    //   latestPartial[key] = partial[key];
                    //   break;
                    default:
                        latestPartial[key] = element[key];
                }
            }
            return latestPartial;
        };
        const applyLatestChangesInternal = (deltas) => {
            const modifiedDeltas = new Map();
            for (const [id, delta] of deltas.entries()) {
                const existingElement = elements.get(id);
                if (existingElement) {
                    const modifiedDelta = Delta.create(delta.deleted, delta.inserted, modifier(existingElement), 'inserted');
                    modifiedDeltas.set(id, modifiedDelta);
                }
                else {
                    modifiedDeltas.set(id, delta);
                }
            }
            return modifiedDeltas;
        };
        const added = applyLatestChangesInternal(this.added);
        const removed = applyLatestChangesInternal(this.removed);
        const updated = applyLatestChangesInternal(this.updated);
        return ElementsChange.create(added, removed, updated, this.api, {
            shouldRedistribute: true, // redistribute the deltas as `isDeleted` could have been updated
        });
    }
    applyTo(elements, snapshot) {
        const nextElements = new Map(elements);
        const flags = {
            containsVisibleDifference: false,
            containsZindexDifference: false,
        };
        const applyDeltas = ElementsChange.createApplier(nextElements, snapshot, flags);
        const addedElements = applyDeltas(this.added);
        const removedElements = applyDeltas(this.removed);
        const updatedElements = applyDeltas(this.updated);
        this.added.forEach((delta, id) => {
            console.log('added', id, delta);
        });
        this.removed.forEach((delta, id) => {
            console.log('removed', id, delta);
        });
        this.updated.forEach((delta, id) => {
            const element = nextElements.get(id);
            if (element) {
                this.api.updateNode(element, delta.inserted);
            }
        });
        return [nextElements, flags.containsVisibleDifference];
    }
    isEmpty() {
        return (this.added.size === 0 &&
            this.removed.size === 0 &&
            this.updated.size === 0);
    }
}
ElementsChange.satisfiesAddition = ({ deleted, inserted, }) => 
// dissallowing added as "deleted", which could cause issues when resolving conflicts
deleted.isDeleted === true && !inserted.isDeleted;
ElementsChange.satisfiesRemoval = ({ deleted, inserted, }) => !deleted.isDeleted && inserted.isDeleted === true;
ElementsChange.satisfiesUpdate = ({ deleted, inserted, }) => !!deleted.isDeleted === !!inserted.isDeleted;
ElementsChange.createGetter = (elements, snapshot, flags) => (id, partial) => {
    let element = elements.get(id);
    if (!element) {
        // always fallback to the local snapshot, in cases when we cannot find the element in the elements array
        element = snapshot.get(id);
        if (element) {
            // as the element was brought from the snapshot, it automatically results in a possible zindex difference
            flags.containsZindexDifference = true;
            // as the element was force deleted, we need to check if adding it back results in a visible change
            if (partial.isDeleted === false ||
                (partial.isDeleted !== true && element.isDeleted === false)) {
                flags.containsVisibleDifference = true;
            }
        }
    }
    return element;
};
ElementsChange.createApplier = (nextElements, snapshot, flags) => {
    const getElement = ElementsChange.createGetter(nextElements, snapshot, flags);
    return (deltas) => Array.from(deltas.entries()).reduce((acc, [id, delta]) => {
        const element = getElement(id, delta.inserted);
        if (element) {
            const newElement = ElementsChange.applyDelta(element, delta, flags);
            nextElements.set(newElement.id, newElement);
            acc.set(newElement.id, newElement);
        }
        return acc;
    }, new Map());
};
// This function tracks updates of text elements for the purposes for collaboration.
// The version is used to compare updates when more than one user is working in
// the same drawing. Note: this will trigger the component to update. Make sure you
// are calling it either from a React event handler or within unstable_batchedUpdates().
export const mutateElement = (entity, element, updates) => {
    let didChange = false;
    for (const key in updates) {
        const value = updates[key];
        if (typeof value !== 'undefined') {
            element[key] = value;
            didChange = true;
        }
    }
    if (!didChange) {
        return element;
    }
    const { name, visibility } = updates;
    const { fill, stroke, strokeWidth, strokeCap, strokeJoin, strokeAlignment, fontSize, width, height, x, y, cx, cy, r, rx, ry, } = updates;
    if (!isNil(name)) {
        entity.write(Name).value = name;
    }
    if (!isNil(visibility)) {
        entity.write(Visibility).value = visibility;
    }
    if (!isNil(fill)) {
        if (isGradient(fill)) {
            removeComponent(entity, FillSolid);
            createOrSetComponent(entity, FillGradient, { value: fill });
        }
        else {
            removeComponent(entity, FillGradient);
            createOrSetComponent(entity, FillSolid, { value: fill });
        }
    }
    if (!isNil(stroke)) {
        createOrSetComponent(entity, Stroke, { color: stroke });
    }
    if (!isNil(strokeWidth)) {
        createOrSetComponent(entity, Stroke, { width: strokeWidth });
    }
    if (!isNil(strokeCap)) {
        createOrSetComponent(entity, Stroke, { linecap: strokeCap });
    }
    if (!isNil(strokeJoin)) {
        createOrSetComponent(entity, Stroke, { linejoin: strokeJoin });
    }
    if (!isNil(strokeAlignment)) {
        createOrSetComponent(entity, Stroke, { alignment: strokeAlignment });
    }
    if (!isNil(fontSize)) {
        entity.write(Text).fontSize = fontSize;
    }
    // TODO: Polyline, Path, Text
    if (!isNil(width)) {
        if (entity.has(Rect)) {
            entity.write(Rect).width = width;
        }
    }
    if (!isNil(height)) {
        if (entity.has(Rect)) {
            entity.write(Rect).height = height;
        }
    }
    if (!isNil(x)) {
        if (entity.has(Rect)) {
            entity.write(Rect).x = x;
        }
    }
    if (!isNil(y)) {
        if (entity.has(Rect)) {
            entity.write(Rect).y = y;
        }
    }
    if (!isNil(cx)) {
        if (entity.has(Circle)) {
            entity.write(Circle).cx = cx;
        }
        else if (entity.has(Ellipse)) {
            entity.write(Ellipse).cx = cx;
        }
    }
    if (!isNil(cy)) {
        if (entity.has(Circle)) {
            entity.write(Circle).cy = cy;
        }
        else if (entity.has(Ellipse)) {
            entity.write(Ellipse).cy = cy;
        }
    }
    if (!isNil(r)) {
        if (entity.has(Circle)) {
            entity.write(Circle).r = r;
        }
    }
    if (!isNil(rx)) {
        if (entity.has(Ellipse)) {
            entity.write(Ellipse).rx = rx;
        }
    }
    if (!isNil(ry)) {
        if (entity.has(Ellipse)) {
            entity.write(Ellipse).ry = ry;
        }
    }
    if (isNil(element.version)) {
        element.version = 0;
    }
    Object.assign(element, updates);
    element.version++;
    element.versionNonce = randomInteger();
    element.updated = getUpdatedTimestamp();
    return element;
};
//# sourceMappingURL=ElementsChange.js.map